import pytest
from gentry.tree import Tree, Visitor, Count


class DummyTree(Tree):
    groups = {"children"}


class DummyVisitor(Visitor):
    def _do_dummyvisitor(self, tree):
        return f"visited:{tree.label}"


def make_simple_tree():
    # root
    # ├── left
    # └── right
    class MyTree(Tree):
        _groups = {"left", "right"}

    left = MyTree(label="left")
    right = MyTree(label="right")
    root = MyTree(label="root")
    root.left.append(left)
    root.right.append(right)
    return root, left, right


def test_visitor_basic():
    root, left, right = make_simple_tree()

    class MyVisitor(Visitor):
        def _do_myvisitor(self, tree):
            return f"visited:{tree.label}"

    v = MyVisitor(root)
    result = v.visit()
    assert "MyTree" in result
    assert result["MyTree"] == "visited:root"
    assert "children" in result
    assert "left" in result["children"]
    assert "right" in result["children"]


def test_visitor_get_visitor_strict():
    root, _, _ = make_simple_tree()

    class MyVisitor(Visitor):
        def _do_myvisitor_MyTree(self, tree):
            return "ok"

    v = MyVisitor(root, strict=True)
    result = v.visit()
    assert result["MyTree"] == "ok"
    left_child_group = result["children"]["left"]
    assert len(left_child_group) == 1
    assert left_child_group[0]["MyTree"] == "ok"
    right_child_group = result["children"]["right"]
    assert len(right_child_group) == 1
    assert right_child_group[0]["MyTree"] == "ok"


def test_visitor_get_visitor_missing():
    root, _, _ = make_simple_tree()

    class MyVisitor(Visitor):
        pass

    v = MyVisitor(root, strict=True)
    with pytest.raises(NotImplementedError):
        v._get_visitor(root)


def test_visit_returns_expected_structure():
    root, left, right = make_simple_tree()

    class MyVisitor(Visitor):
        def _do_myvisitor(self, tree):
            return tree.label

    v = MyVisitor(root)
    result = v.visit()
    assert result["MyTree"] == "root"
    assert result["children"]["left"][0]["MyTree"] == "left"
    assert result["children"]["right"][0]["MyTree"] == "right"


def test_count_simple_tree():
    root, left, right = make_simple_tree()
    c = Count(root)
    assert c.count() == 3  # root + left + right


def test_count_nested_tree():
    class MyTree(Tree):
        groups = {"children"}

    leaf1 = MyTree(label="leaf1")
    leaf2 = MyTree(label="leaf2")
    mid = MyTree(label="mid", children={"children": [leaf1, leaf2]})
    root = MyTree(label="root", children={"children": [mid]})
    c = Count(root)
    assert c.count() == 4  # root, mid, leaf1, leaf2


def test_count_empty_tree():
    t = Tree(label="root")
    c = Count(t)
    assert c.count() == 1


def test_count_with_subclass():
    class MyTree(Tree):
        groups = {"children"}

    t = MyTree(label="root")
    c = Count(t)
    assert c.count() == 1


def test_count_sum_handles_various_types():
    # Test _sum with int, list, dict
    assert Count._sum(5) == 5
    assert Count._sum([1, 2, 3]) == 6
    assert Count._sum({"a": 1, "b": 2}) == 3
    nested = {"a": [1, 2], "b": {"c": 3}}
    assert Count._sum(nested) == 6



class HTMLLayout:
    """
    A mixin class for Tree that adds a __str__ method that will render a node and its children as an svg.

    Assumes the existance of the following attributes:
      _children     a dict[str,list]
      label         a str
      properties    a dict
    """

    _include_properties = False

    def __init__(
        self,
        include_properties=None,
    ) -> None:
        self._iinclude_properties = include_properties


    def __str__(self) -> str:
        html = self._box()
        prolog = '''<html>
        <head><title>Tree</title></head>
        <style>
        .column, .group, .outercontainer, .parent, .leaf, .properties { display:flex; flex-direction:column;}
        .children, .groupitems, .property { display: flex; flex-direction:row; align-items:stretch;}
        .children {border-top: black solid 1pt;}
        .parent {border-left: black solid 1pt; font-weight:bold;}
        .outercontainer { width:fit-content;}
        .leaf, .parent {padding: 0.5em;}
        .leaf {font-weight: bold;}
        .leaf:last-of-type { padding-right: 1em;} 
        .groupname {padding: 0.2em; font-size:8pt;}
        .group { background: #b8b8b8; background: linear-gradient(90deg, rgb(230 230 230) 0%, rgb(237 237 237) 25%, rgba(255, 255, 255, 1) 100%);}
        .key { padding-right:1em; }
        </style>
        <body><div class="outercontainer"
        '''
        epilog = """</div></body>
        </html>"""
        return f"{prolog}{html}{epilog}"
    
    def _box(self):
        node_name = self.label

        include_properties = self._include_properties  # class var
        if (
            self._iinclude_properties is not None
        ):  # override if instance variable is not None
            include_properties = self._iinclude_properties
        if include_properties:  # neither None or False
            props = [f'<div class="property"><div class="key">{k}</div><div class="value">{v}</div></div>' for k, v in self.properties.items()]
            props = f'<div class="properties">{"".join(props)}</div>'
        else:
            props = ""

        if self.is_leaf():
            return f'<div class="leaf"><div class="nodename">{node_name}</div>{props}</div>\n'
        else:
            groups = {}
            for group, children in self._children.items():
                childitems = []
                for child in children:
                    if child is not None:
                        childitems.append(child._box())
                groups[group] = "".join(childitems)
            groupdivs = "".join(f'<div class="group">\n<div class="groupname">{group}</div>\n<div class="groupitems">{html}</div>\n</div>\n' for group,html in groups.items())
            return f'<div class="column">\n<div class="parent"><div class="nodename">{node_name}</div>{props}</div>\n<div class="children">{groupdivs}</div>\n</div>'



import pytest
from collections import defaultdict
from gentry.tree import Tree


class TestTree:
    def test_init_default_children(self):
        t = Tree(label="root")
        assert t.label == "root"
        assert isinstance(t._children, defaultdict)
        assert t._children == {}

    def test_init_with_children_dict(self):
        child = Tree(label="child")
        children = {"group1": [child]}
        t = Tree(label="root", children=children)
        assert t._children["group1"] == [child]

    def test_init_with_children_defaultdict(self):
        child = Tree(label="child")
        children = defaultdict(list, {"group1": [child]})
        t = Tree(label="root", children=children)
        assert t._children["group1"] == [child]

    def test_getattr_group(self):
        class MyTree(Tree):
            _groups = {"group1"}

        t = MyTree(label="root")
        c = MyTree(label="child")
        t.group1.append(c)
        assert t.group1 == [c]

    def test_getattr_non_group_raises(self):
        t = Tree(label="root")
        with pytest.raises(AttributeError):
            _ = t.nonexistent

    def test_setattr_group(self):
        class MyTree(Tree):
            _groups = {"group1"}

        t = MyTree(label="root")
        c = MyTree(label="child")
        t.group1 = [c]
        assert t._children["group1"] == [c]

    def test_setattr_group_non_list_raises(self):
        class MyTree(Tree):
            _groups = {"group1"}

        t = MyTree(label="root")
        with pytest.raises(AttributeError):
            t.group1 = "notalist"

    def test_setattr_normal(self):
        t = Tree(label="root")
        t.foo = 123
        assert t.foo == 123

    def test_repr(self):
        t = Tree(label="root")
        r = repr(t)
        assert "Tree" in r and "label=root" in r

    def test_groups_inheritance(self):
        class A(Tree):
            _groups = {"left", "right"}

        class B(A):
            pass

        bnode = B(label="b")
        lnode = B(label="left")
        rnode = B(label="right")
        bnode.left.append(lnode)
        bnode.right = [rnode]
        assert bnode.left == [lnode]
        assert bnode.right == [rnode]

    def test_subclass_with_own_groups(self):
        class Base(Tree):
            _groups = {"foo"}

        class Sub(Base):
            _groups = {"bar"}

        s = Sub(label="s")
        b = Sub(label="bar")
        s.bar.append(b)
        assert s.bar == [b]
        # foo should not be a group in Sub
        with pytest.raises(AttributeError):
            _ = s.foo

    def test_subclass_invalid_groups_type(self):
        with pytest.raises(AttributeError):

            class Bad(Tree):
                _groups = ["not", "a", "set"]

    def test_subclass_groups_with_invalid_names(self):
        with pytest.raises(AttributeError):

            class Bad(Tree):
                _groups = {"_bad"}

        with pytest.raises(AttributeError):

            class Bad2(Tree):
                _groups = {"not-valid!"}

        with pytest.raises(AttributeError):

            class Bad3(Tree):
                _groups = {123}

    def test_subclass_groups_are_independent(self):
        class A(Tree):
            _groups = {"left"}

        class B(Tree):
            _groups = {"right"}

        a = A(label="a")
        b = B(label="b")
        lnode = A(label="left")
        rnode = B(label="right")
        a.left.append(lnode)
        b.right.append(rnode)
        assert a.left == [lnode]
        assert b.right == [rnode]
        with pytest.raises(AttributeError):
            _ = a.right
        with pytest.raises(AttributeError):
            _ = b.left
        assert b.right == [rnode]

    def test_is_leaf(self):
        class MyTree(Tree):
            _groups = {"kids"}

        # A node with no children should be a leaf
        t = MyTree(label="root")
        assert t.is_leaf()

        # A node with one child in a group shouldn't be a leaf
        child = MyTree(label="child")
        t.kids.append(child)
        assert not t.is_leaf()

        # A node with multiple children in multiple groups
        t2 = MyTree(label="root2")
        t2.kids.append(MyTree(label="c1"))
        t2.kids.append(MyTree(label="c2"))
        assert not t2.is_leaf()

    def test_initialization_with_keyword_argument(self):
        class A(Tree):
            _groups = {"agroup"}

        child = A(label="child")
        a = A(label="a", agroup=[child])

        assert a.agroup == [child]

    def test_initialization_with_unknown_keyword_argument(self):
        class A(Tree):
            _groups = {"agroup"}

        child = A(label="child")
        with pytest.raises(TypeError):
            _ = A(label="a", unknown=[child])

    def test_initialization_with_duplicated_keyword_argument(self):
        class A(Tree):
            _groups = {"agroup"}

        child = A(label="child")
        with pytest.raises(ValueError):
            _ = A(label="a", agroup=[child], children={"agroup":[child]})

    def test_initialization_with_mixed_group_argument(self):
        class A(Tree):
            _groups = {"agroup", "another_group"}

        child1 = A(label="child1")
        child2 = A(label="child2")

        a = A(label="a", agroup=[child1], children={"another_group":[child2]})
        assert a.agroup == [child1]
        assert a.another_group == [child2]



# just an example on how to use the Tree class

from collections import defaultdict
from .tree import Tree, Count
from .mermaid import Mermaid, Shape, Style

class Family(Tree, Mermaid):...

class Person(Tree, Mermaid):
    _groups = { "children"}
    
class GrandMother(Person):
    _style = Style.function
    _shape = Shape.braces

class Mother(Person):
    _groups = {"girls", "boys"}
 
class Child(Person):
    _include_properties = True

class FamilyCount(Count):
    """
    We only count real persons, so we define a specialized member function
    for any Family node that returns 0 and will not be counted.
    """
    def _do_count_Family(self, tree:Family):
        return 0

# define a few children  
c1 = Child("Alice")
c2 = Child("Bob")
c3 = Child("Cherryl", properties={"chess master": "ELO 2235"})
c4 = Child("Dick")
c5 = Child("Ellen")
c6 = Child("Fergal")
c7 = Child("Gladys", properties={"drivers license": 2023, "nose piercing": 2024})
c8 = Child("Hank")

# Mothers have girls and boys attributes defined, so those can be assigned directly
m1 = Mother("Anna")
m1.girls = [c1,c3]
m1.boys = [c2,c4]

m2 = Mother("Beatrice")
m2.girls.append(c5)     # under the hood they are all items in a defaultdict(list) so we can append directly
m2.girls.append(c7)
m2.boys = [c6,c8]

# Grandmothers do not make the distinction
g1 = GrandMother("Granny")
g1.children = [m1, m2]
# so this would fail:
# g1.girls = [m1, m2]

# Family does not have any direct access attributes defined, but anything
# that is passed as the children argument (and can be converted to a defaultdict(list))
# will be added to the _children attribute, so will still be automatically discovered
# by Visitor derived classes.
f1 = Family("The Andersons", children={"matriarch":[g1]})

counter = FamilyCount(f1, strict=False)

assert counter.count() == 11

print(f1)
